"""
 Created on May 19, 2018
 @author: Lorenzo Delgado <lorenzo.delgado@lnsd.es>
"""
from nssm.abstract.collections import AbstractEnum

__all__ = ["StartupType", "PriorityLevel", "ServiceType", "ExitAction"]


class StartupType(AbstractEnum):
        """
        AUTOMATIC: Automatic startup at boot.
        DELAYED: Delayed startup at boot.
        MANUAL: Manual startup.
        DISABLED: Service is disabled.

        Note that SERVICE_DELAYED_START is not supported on versions of
        Windows prior to Vista. `nssm` will set the service to automatic
        startup if delayed start is unavailable.
        """
        AUTOMATIC = "SERVICE_AUTO_START"
        DELAYED = "SERVICE_DELAYED_AUTO_START"
        MANUAL = "SERVICE_DEMAND_START"
        DISABLED = "SERVICE_DISABLED"


class PriorityLevel(AbstractEnum):
    """
    The AppPriority parameter takes a priority class constant as specified
    in the SetPriorityClass() documentation.
    """
    REALTIME = "REALTIME_PRIORITY_CLASS"
    HIGH = "HIGH_PRIORITY_CLASS"
    ABOVE_NORMAL = "ABOVE_NORMAL_PRIORITY_CLASS"
    NORMAL = "NORMAL_PRIORITY_CLASS"
    BELOW_NORMAL = "BELOW_NORMAL_PRIORITY_CLASS"
    IDLE = "IDLE_PRIORITY_CLASS"


class ServiceType(AbstractEnum):
    """
    The Type parameter is used to query or set the service type. `nssm`
    recognises all currently documented service types but will only allow
    setting one of two types: SERVICE_WIN32_OWN_PROCESS or
    SERVICE_INTERACTIVE_PROCESS
    """
    # A standalone service. This is the default.
    STANDALONE = "SERVICE_WIN32_OWN_PROCESS"
    # A service which can interact with the desktop.
    DESKTOP = "SERVICE_INTERACTIVE_PROCESS"


class ExitAction(AbstractEnum):
    """
    Exit action
    """
    RESTART = "Restart"
    IGNORE = "Ignore"
    EXIT = "Exit"
    SUICIDE = "Suicide"


PARAM_MAP = {
    # Application
    "path": "Application",
    "startup_dir": "AppDirectory",
    "arguments": "AppParameters",

    # Details
    "display_name": "DisplayName",
    "description": "Description",
    "startup": "Start",

    # Log on
    "user_account": "ObjectName",
    "type": "Type",

    # Dependencies
    "dependencies": "DependsOnService",

    # Process
    "process_priority": "AppPriority",
    "console_window": "AppNoConsole",
    "cpu_affinity": "AppAffinity",

    # Shutdown
    "terminate_process": "AppStopMethodSkip",
    "stop_console": "AppStopMethodConsole",
    "stop_window": "AppStopMethodWindow",
    "stop_threads": "AppStopMethodThreads",

    # Exit action
    "restart_throttling": "AppThrottle",
    "restart_delay": "AppRestartDelay",
    "action_on_exit": "AppExit",

    # I/O
    "stdout": "AppStdout",
    "stderr": "AppStderr",

    # File rotation
    "rotate_files": "AppRotateFiles",
    "rotate_online": "AppRotateOnline",
    "stdout_creation_disposition": "AppStdoutCreationDisposition",
    "stderr_creation_disposition": "AppStderrCreationDisposition",
    "rotation_time": "AppRotateSeconds",
    "rotation_size": "AppRotateBytes",

    # Environment
    "env": "AppEnvironmentExtra"
}



"""
 Created on May 19, 2018
 @author: Lorenzo Delgado <lorenzo.delgado@lnsd.es>
"""


class NssmException(Exception):
    """

    """
    MESSAGE_TEMPLATE = "[{service}] {message}"
    DEFAULT_MESSAGE = "NSSM error occurred"

    def __init__(self, service, msg=None):
        """

        """
        if msg is None:
            msg = self.DEFAULT_MESSAGE

        msg = self.MESSAGE_TEMPLATE.format(service=service, message=msg)
        super(Exception, self).__init__(msg)


class ServiceInstallException(NssmException):
    """

    """
    DEFAULT_MESSAGE = "Service installation failed"


class ServiceRemoveException(NssmException):
    """

    """
    DEFAULT_MESSAGE = "Service removal failed"


class ServiceStartException(NssmException):
    """

    """
    DEFAULT_MESSAGE = "Service start failed"


class ServiceStopException(NssmException):
    """

    """
    DEFAULT_MESSAGE = "Service stop failed"


class ServicePauseException(NssmException):
    """

    """
    DEFAULT_MESSAGE = "Service pause failed"


class ServiceResumeException(NssmException):
    """

    """
    DEFAULT_MESSAGE = "Service resume failed"


class ServiceConfigurationException(NssmException):
    """

    """
    DEFAULT_MESSAGE = "Service configuration failed"



# -*- coding: utf-8 -*-
"""
 Python wrapper for NSSM

 Created on May 19, 2018
 @author: Lorenzo Delgado <lorenzo.delgado@lnsd.es>
"""

from nssm import metadata


__version__ = metadata.version
__author__ = metadata.authors[0]
__license__ = metadata.license
__copyright__ = metadata.copyright



"""
 Created on May 19, 2018
 @author: Lorenzo Delgado <lorenzo.delgado@lnsd.es>
"""
from voluptuous import Schema, Required, Optional, Or, Coerce, \
                       humanize as hum, REMOVE_EXTRA, ALLOW_EXTRA

from nssm.abstract.collections import AbstractDict
from nssm.parameters import StartupType, PriorityLevel, ServiceType, ExitAction


class ServiceConfiguration(AbstractDict):
    """
    Service configuration object
    """

    def __init__(self, *args, **kwargs):
        """
        Service configuration class constructor
        """
        super(ServiceConfiguration, self).__init__()

        schema = self._get_schema()
        self.dictionary = hum.validate_with_humanized_errors(kwargs, schema)

    @staticmethod
    def _get_schema():
        """
        Build a voluptuous schema to validate the input keyword arguments
        :return: Voluptuous schema
        :rtype: :class:`Schema`
        """
        return Schema({
            # Application
            Optional("path"): str,
            Optional("startup_dir"): str,
            Optional("arguments"): Or(str, list),

            # Details
            Optional("display_name"): str,
            Optional("description"): str,
            Optional("startup"): Coerce(StartupType.coerce),

            # Log on
            Optional("user_account"): Or(str, Schema({
                Required("username"): str,
                Required("password"): str
            }, extra=REMOVE_EXTRA)),
            Optional("type"): Coerce(ServiceType.coerce),

            # Dependencies
            Optional("dependencies"): Or(str, list),

            # Process
            Optional("process_priority"): Coerce(PriorityLevel.coerce),
            Optional("console_window"): bool,
            Optional("cpu_affinity"): Or("All", int),

            # Shutdown
            Optional("terminate_process"): bool,
            Optional("stop_console"): int,
            Optional("stop_window"): int,
            Optional("stop_threads"): int,

            # Exit action
            Optional("restart_throttling"): int,
            Optional("restart_delay"): int,
            Optional("action_on_exit"): Schema({
                Required("Default"): Coerce(ExitAction.coerce)
            }, extra=ALLOW_EXTRA),

            # I/O
            Optional("stdout"): str,
            Optional("stderr"): str,

            # File rotation
            Optional("rotate_files"): bool,
            Optional("rotate_online"): bool,
            Optional("stdout_creation_disposition"): int,
            Optional("stderr_creation_disposition"): int,
            Optional("rotation_time"): int,
            Optional("rotation_size"): int,

            # Environment
            Optional("env"): Or(str, dict)
        }, extra=REMOVE_EXTRA)



"""
 Created on May 19, 2018
 @author: Lorenzo Delgado <lorenzo.delgado@lnsd.es>
"""
from nssm.abstract.collections import AbstractEnum

from nssm.wrapper import Wrapper
from nssm.parameters import PARAM_MAP
from nssm.configuration import ServiceConfiguration


class ServiceStatus(AbstractEnum):
    STOPPED = "SERVICE_STOPPED"
    RUNNING = "SERVICE_RUNNING"
    PAUSED = "SERVICE_PAUSED"


class Service(object):
    """
    NSSM Service class
    """

    def __init__(self, name, path, **kwargs):
        """
        :param name: Service name
        :type name: str
        :param path: Executable path
        :type path: str
        """
        self.name = name
        self.path = path
        self.configuration = ServiceConfiguration(**kwargs)

    def install(self):
        """
        Install the service
        """
        Wrapper.command("install", self.name, self.path)

        if self.configuration:
            self.configure(self.configuration)

    def remove(self):
        """
        Remove the service
        """
        Wrapper.command("remove", self.name, "confirm")

    def start(self):
        """
        Start the service
        """
        Wrapper.command("start", self.name)

    def stop(self):
        """
        Stop the service
        """
        Wrapper.command("stop", self.name)

    def restart(self):
        """
        Restart the service
        """
        Wrapper.command("restart", self.name)

    def pause(self):
        """
        Pause the service
        """
        Wrapper.command("pause", self.name)

    def resume(self):
        """
        Resume the service
        """
        Wrapper.command("continue", self.name)

    def rotate(self):
        """
        Triggers on-demand rotation for `nssm` services with I/O redirection
        and online rotation enabled.

        `nssm` accepts user-defined control 128 as a cue to begin output
        file rotation. Non-nssm services might respond to control 128 in
        their own way (or ignore it, or crash).
        """
        Wrapper.command("rotate", self.name)

    def status(self):
        """
        Query the service's status
        :return: Service status
        :rtype: :class:`ServiceStatus`
        """
        rc, out = Wrapper.command("status", self.name)
        return ServiceStatus(out.strip())

    def configure(self, config):
        """
        Configure the service
        :param config: Service configuration
        :type config: :class:`ServiceConfiguration`
        """
        msg = "Unknown configuration object: {}"
        assert isinstance(config, ServiceConfiguration), msg.format(config)

        # Update service configuration
        self.configuration.update(config)

        for param, value in config.items():

            if param == "user_account" and isinstance(value, dict):
                value = value["username"] + " " + value["password"]

            elif param == "env" and isinstance(value, dict):
                env = ""
                for k, v in value.items():
                    env += k + "=" + v
                value = env

            elif param == "action_on_exit":
                for k, v in value.items():
                    v = v.value
                    Wrapper.command("set", self.name, PARAM_MAP[param], k, v)
                continue

            if isinstance(value, AbstractEnum):
                value = value.value
            elif isinstance(value, bool):
                value = 1 if value else 0

            Wrapper.command("set", self.name, PARAM_MAP[param], value)





from enum import Enum


class AbstractDict(object):
    """
    This is an abstract class which implements an dictionary like object with
    some additional features.
    """
    __marker = object()

    def __init__(self):
        self.dictionary = {}

    def __getitem__(self, key):
        return self.dictionary[self._keytransformer(key)]

    def __setitem__(self, key, value):
        self.dictionary[self._keytransformer(key)] = value

    def __delitem__(self, key):
        del self.dictionary[self._keytransformer(key)]

    def __iter__(self):
        return iter(self.dictionary)

    def __len__(self):
        return len(self.dictionary)

    def __str__(self):
        return str(self.dictionary)

    def __repr__(self):
        return str(self.dictionary)

    def update(self, d):
        self._recursive_update(self.dictionary, d)

    def iterkeys(self):
        return iter(self)

    def iteritems(self):
        for key in self:
            yield (key, self[key])

    def itervalues(self):
        for key in self:
            yield self[key]

    def keys(self):
        return self.dictionary.keys()

    def items(self):
        return [(key, self[key]) for key in self]

    def values(self):
        return [self[key] for key in self]

    def pop(self, key, default=__marker):
        try:
            value = self[key]
        except KeyError:
            if default is self.__marker:
                raise
            return default
        else:
            del self[key]
            return value

    def popitem(self):
        try:
            key = next(iter(self))
        except StopIteration:
            raise KeyError
        value = self[key]
        del self[key]
        return key, value

    def clear(self):
        try:
            while True:
                self.popitem()
        except KeyError:
            pass

    def setdefault(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            self[key] = default
        return default

    def pretty_print(self):
        return self._recursive_print(self.dictionary, 0)

    def _recursive_update(self, d1, d2):
        for k, v in sorted(d2.items()):
            if k in d1.keys() and type(v) == dict:
                self._recursive_update(d1[k], v)
            else:
                d1[k] = v

    def _recursive_print(self, d, nesting):
        result = ""

        for k, v in sorted(d.items()):
            if type(v) == dict:
                continue

            result += " "*nesting + "%-30s" % (str(k),) + ": " + str(v) + "\n"

        for k, v in sorted(d.items()):
            if type(v) != dict:
                continue

            result += " "*nesting + "%s" % (str(k),) + "...\n"
            result += self._recursive_print(d[k], nesting+4)

        return result

    def _keytransformer(self, key):
        """When overridden permits manipulation of the key before accessing
        the dictionary"""
        return key


class AbstractEnum(Enum):
    """
    Python has built-in support for Enum class types from 3.4 on, backport
    is provided in enum34 package. This class further extends this Enum class
    with an easy method to get the defined symbolic enumerations.
    """

    @classmethod
    def coerce(cls, val):
        """
        Return the desired enum object or throw an error, the enum object will
        be searched based on its name or value.

        :param val: the value or name of the desired enum
        :return: the enum object
        """
        try:
            return cls[val]
        except KeyError:
            try:
                return cls(val)
            except ValueError:
                msg = '"{0}" not a valid enum value or name of {1}, possible' \
                      ' names are {2} and possible values are {3}.'
                raise ValueError(msg.format(val, cls.__name__,
                                            ','.join(cls.get_names()),
                                            str(cls.get_values())))

    @classmethod
    def get_names(cls):
        """
        Returns all the names of the symbolic enumerations defined. Including
        the aliases.

        :return: list of all enum names
        """
        names = []
        for name, member in cls.__members__.items():
            names.append(name)
        return names

    @classmethod
    def get_values(cls):
        """
        Returns all the values of the defined symbolic enumerations.

        :return: list of values
        """
        values = []
        for member in list(cls):
            values.append(member.value)
        return values



# -*- coding: utf-8 -*-
"""
 Project metadata
 Information describing the project.

 Created on May 19, 2018
 @author: Lorenzo Delgado <lorenzo.delgado@lnsd.es>
"""

# The package name, which is also the "UNIX name" for the project.
package = 'pynssm'
project = "PyNSSM"
project_no_spaces = project.replace(' ', '')
version = '2.24.1-alpha.2'
description = 'Python wrapper for NSSM - the Non-Sucking Service Manager (nssm.cc)'
authors = ['Lorenzo Delgado']
authors_string = ', '.join(authors)
emails = ['lorenzo.delgado@lnsd.es']
license = 'MIT'
copyright = '2018 ' + authors_string
url = 'https://github.com/LNSD/pynssm'
download_url = 'https://github.com/LNSD/pynssm/archive/2.24.1-alpha.2.tar.gz'



from nssm.service import Service, ServiceConfiguration
from nssm.exceptions import *

config = ServiceConfiguration(
    arguments=r"C:\Users\ldelgado\projects\MATS\webgui\pynssm\tests\dummy.py",
    startup_dir=r"C:\MoCA\dummy",
    display_name="Test Service",
    description="This is a demo service",
    stdout=r"C:\MoCA\dummy\test_service_stdout.log",
    stderr=r"C:\MoCA\dummy\test_service_stderr.log"
)

service = Service("test_service",
                  r"C:\Users\ldelgado\Envs\nssm\Scripts\python.exe")

try:
    service.remove()
except ServiceRemoveException as err:
    print err.message

# try:
#     service.install()
# except ServiceInstallException:
#     print err.message
#
# service.configure(config)
# service.start()
# print service.status()
# service.stop()
# print service.status()



"""
 Created on May 19, 2018
 @author: Lorenzo Delgado <lorenzo.delgado@lnsd.es>
"""
import os
import re
import subprocess as sp
import logging

from nssm.exceptions import *

log = logging.getLogger(__name__)


class Wrapper(object):

    @staticmethod
    def command(command, service_name, *args):
        """
        Service installation
        :param command: nssm command
        :type command: str
        :param service_name: Service name
        :type service_name: str
        :param args: Program arguments
        :type args: str or tuple
        :return: Return code and output of the command
        :rtype: (int, str)
        """
        # Build de command string
        cmd = [Wrapper._nssm_exe(), command, service_name]
        cmd += args if len(args) == 1 else list(args)

        # Format in an unique command string
        cmd = " ".join(cmd)

        # Print command string for debugging purpose
        log.debug(cmd)

        # Run the command
        try:
            out = sp.check_output(cmd, stderr=sp.STDOUT,
                                  universal_newlines=True)
            return 0, out.decode("utf-16")

        except sp.CalledProcessError as err:
            msg = "{msg} ({rc}): {out}"
            err.output = err.output.decode("utf-16")
            err.output = re.sub(r"\n+", " ", err.output)

            command_except_map = {
                "install": ServiceInstallException,
                "remove": ServiceRemoveException,
                "start": ServiceStartException,
                "restart": ServiceStartException,
                "stop": ServiceStopException,
                "pause": ServicePauseException,
                "continue": ServiceResumeException
            }

            if command in command_except_map.keys():
                msg = msg.format(
                    msg=command_except_map[command].DEFAULT_MESSAGE,
                    rc=err.returncode,
                    out=err.output
                )
                raise command_except_map[command](service_name, msg)

            else:
                msg = msg.format(msg=NssmException.DEFAULT_MESSAGE,
                                 rc=err.returncode,
                                 out=err.output)
                raise NssmException(service_name, msg)

    @staticmethod
    def _nssm_exe():
        """
        Build executable full path depending on processor architecture
        :return: Path to the bundled executable
        :rtype: str
        """
        arch = "win64" if os.environ['PROCESSOR_ARCHITECTURE'] else "win32"
        return os.path.join(os.path.dirname(__file__), "bin", arch, "nssm.exe")
